/*
  RabidVWM LiteStep Module - version 1.0
	Window Management Functions

  Copyright (C) 2002 Joshua Seagoe
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


#include "RabidVWM.h"
#include <stdlib.h>

//#define TRACE_UPDATE

// I have the feeling I should find a way to use LiteStep's
// winlist.dll, but it's not documented so I don't know
// (actually, that probly means it's internal confidencial and I shouldn't use it)
int InsertWindow(int position,HWND hWnd);
void DeleteWindow(int position);


// return true if the window is hung
// uses undocumented calls which differ on nt/9x
bool IsWndHung(HWND hWnd)
{
	if (dIsHungAppWindow) return dIsHungAppWindow(hWnd)!=FALSE;                          /* nt */ 
	if (dIsHungThread) return dIsHungThread(GetWindowThreadProcessId(hWnd,NULL))!=FALSE; /* 9x */ 
  return true; /* should never happen, but prefer being slow to allowing ls to freeze */ 
}

// return true if the window is minimized
// (required because IsIconic doesn't seem to do the trick)
bool IsWndMinimized(const s_WinData *pData)
{
	ASSERT(READABLE(pData,sizeof(s_WinData)));
	return (
		/* hardcoding coordinates is bad, i know, but what'cha gonna do? */ 
		/* I don't know exactly *who* moves them there.                  */ 
		(pData->Area.left==-32000 && pData->Area.top==-32000)
		/* I don't think this ever works, but just in case.              */ 
		|| IsIconic(pData->hWnd));
}

// return true if the window is on a monitor that I'm switching
bool IsOnOwnedMonitor(const s_WinData *pData)
{
	ASSERT(READABLE(pData,sizeof(s_WinData)));
	if (GlobalData.hMonitorToSwitch==NULL) return true;
	if (pData->hMonitor==GlobalData.hMonitorToSwitch) return true;
	return false;
}

// return true if the window will have a miniwindow
bool ShouldDrawWnd(const s_WinData *pData)
{
	ASSERT(READABLE(pData,sizeof(s_WinData)));
	if (!IsOnOwnedMonitor(pData)) return false;
#ifndef SHOW_BERSIRC_ON_ALL_DESKTOPS_TASK_LIST
  if (pData->Mini.right-pData->Mini.left == 0
    ||pData->Mini.bottom-pData->Mini.top == 0)
    return false;
#endif
	return !GlobalData.bNoWindows /* no windows are shown */ 
		&& !IsWndMinimized(pData)
		&& !IsToolWindow(pData->hWnd)
		&& !pData->bSticky;
}

// return true if the window moves when switching desktops
bool ShouldMoveWnd(const s_WinData *pData)
{
	ASSERT(READABLE(pData,sizeof(s_WinData)));
	if (!IsOnOwnedMonitor(pData)) return false;
	return
    /* make it so that this'll help with XP Styles */ 
		/*!(GlobalData.bHideTaskOnSwitch && !GlobalData.bHideTaskAndMove)
		&& 
    */
    !IsToolWindow(pData->hWnd)
		&& !pData->bFix
		/* minimized windows have a fixed position */ 
		&& !IsWndMinimized(pData)
		;
}

// return true if the window should be kept with its parent at all times
bool ShouldFixWnd(const s_WinData *pData)
{
  return
		(IsToolWindow(pData->hWnd) && GetParent(pData->hWnd))
		|| pData->bFix
    ;
}

// return true if the window should only appear on its desktop's task list
bool ShouldHideWnd(const s_WinData *pData)
{
	ASSERT(READABLE(pData,sizeof(s_WinData)));
	if (!IsOnOwnedMonitor(pData)) return false;
	return 
		(GlobalData.bHideTaskOnSwitch
			/* keep minimized windows on all desks             */ 
			/* fixes minimized windows vanishing with HideTask */ 
			&& !IsWndMinimized(pData)
		)
		/* floating toolboxes */ 
		|| ShouldFixWnd(pData)
		;
}

// return true if the window should not be added to
// the known window list if it isn't already there
// (ie, should not be drawn)
static bool ShouldntAddWindow(HWND hWnd)
{
	/* for now, ignore all hidden windows and windows with the magic value */ 
	return !IsWindowVisible(hWnd)
		|| IsLiteStepWnd(hWnd)
		|| IsIconic(hWnd);
// can't ignore these altogether
//		|| (GetWindowLong(hWnd, GWL_EXSTYLE) & WS_EX_TOOLWINDOW);
// Don't ask me what those are, I added that comment months ago.
}

// update the window rectangles
void UpdateRect(s_WinData *pData)
{
	ASSERTR(READABLE(pData,sizeof(s_WinData)),;);
	ASSERTR(WRITEABLE(pData,sizeof(s_WinData)),;);

	/* update the current rect.                 */ 
	GetWindowRect(pData->hWnd,&pData->Area);
#if !defined(NDEBUG) && defined(TRACE_FILE) && defined(TRACE_UPDATE)
	/* to make sure it works in debug mode without flooding the logs */ 
	if (bInLocalInit) {
    TRACE(TRACE_STATUS,"-- UpdateRectN");
	}
#endif
	UpdateRectN(pData);

	/* Tried using GetInternalWindowPos in case */ 
	/* it was faster.  That didn't work right.  */ 
}

// calculate which desk this window is on, in cases where the HomeDesk may not be valid
int GetHomeFromArea(const s_WinData *pData)
{
	ASSERTR(READABLE(pData,sizeof(s_WinData)),0);
	int home;
	/* set home desk from center of rect */ 
	home = DeskFromPoint(
		(pData->Area.left+pData->Area.right)/2,
		(pData->Area.top+pData->Area.bottom)/2);
	/* if that's lost, try the top corners (down a smidge) */ 
	if (home<0) home = DeskFromPoint(pData->Area.left,pData->Area.top+16); /* 16px=1 smidge */ 
	if (home<0) home = DeskFromPoint(pData->Area.right,pData->Area.top+16);

	CWTRACE(TRACE_WARNING,home<0,pData,"Couldn't get home from area.");

	return home;
}

// figure out which monitor a window is on, no matter what desktop it's on
HMONITOR GetWindowMonitor(const s_WinData *pData)
{
	ASSERTR(READABLE(pData,sizeof(s_WinData)),NULL);
  ASSERTV(GlobalData.ScreenSize.cx!=0);
  ASSERTV(GlobalData.ScreenSize.cy!=0);

	/* Can't use MonitorFromWindow because the window won't  */ 
	/* be on any monitor unless it's on the current desktop. */ 
	POINT pt;

	/* get center                 */ 
	pt.x = (pData->Area.left+pData->Area.right)/2;
	pt.y = (pData->Area.top+pData->Area.bottom)/2;

	/* get position modulus desks */ 
	pt.x += GlobalData.ScreenOffset.x;
	pt.y += GlobalData.ScreenOffset.y;

	pt.x %= GlobalData.ScreenSize.cx;
	pt.y %= GlobalData.ScreenSize.cy;

	if (pt.x<0) pt.x += GlobalData.ScreenSize.cx;
	if (pt.y<0) pt.y += GlobalData.ScreenSize.cy;

	pt.x -= GlobalData.ScreenOffset.x;
	pt.y -= GlobalData.ScreenOffset.y;

	/* find what monitor it's on  */ 
	HMONITOR m = NULL;
  if (dMonitorFromPoint) m = dMonitorFromPoint(pt,MONITOR_DEFAULTTONULL);
	CWTRACE(TRACE_WARNING,m==NULL,pData,"Window not on any monitor.");

	return m;
}

// update from "fake" window rect
void UpdateRectN(s_WinData *pData)
{
	int home;
	ASSERTR(READABLE(pData,sizeof(s_WinData)),;);
	ASSERTR(WRITEABLE(pData,sizeof(s_WinData)),;);

	if (GlobalData.bScaleChanged)
	{
    /* if the desktop->monitor scale has      */ 
    /* changed, we need to recalc everything. */ 
		pData->bPosChanged = true;
		pData->bSizeChanged = true;
	}
	else
	{
    /* otherwise, see what has changed in the */ 
    /* window's real area.                    */ 

		if (pData->Area.top != pData->OldArea.top
			||pData->Area.left != pData->OldArea.left)
			pData->bPosChanged = true;

		if ((pData->Area.right-pData->Area.left != pData->OldArea.right-pData->OldArea.left)
			||(pData->Area.bottom-pData->Area.top != pData->OldArea.bottom-pData->OldArea.top))
			pData->bSizeChanged = true;
	}

	/* Reset its home desk and VWM rect if size/position has changed. */ 
	if (pData->bPosChanged || pData->bSizeChanged)
	{
		if (IsWndMinimized(pData))
		{
#if !defined(NDEBUG) && defined(TRACE_FILE) && defined(TRACE_UPDATE)
	/* to make sure it works in debug mode without flooding the logs */ 
	if (bInLocalInit) {
    TRACE(TRACE_STATUS,"-- minimized");
	}
#endif
			/* Minimized windows should always restore to the current desk.     */ 
			/* If they aren't handled specially, they will be detected as lost, */ 
			/* which has some bad side effects.                                 */ 
			SetHomeDesk(pData,GlobalData.CurrentDesk);
		}
		else 
		{
			HWND parent;
			s_WinData *pParent = NULL;

			/* Take a snapshot after the window has had a chance to repaint.    */ 
			GlobalData.bDelaySnapshot = true;

			/* If the window doesn't move with its desk...                      */ 
			if (!ShouldMoveWnd(pData))
			{
				/* set home desk from parent, if any */ 
				parent = GetParent(pData->hWnd);
				pParent = GetWindowPtr(parent);
			}

#if !defined(NDEBUG) && defined(TRACE_FILE) && defined(TRACE_UPDATE)
	/* to make sure it works in debug mode without flooding the logs */ 
	if (bInLocalInit) {
    TRACE(TRACE_STATUS,"-- setting home");
	}
#endif
			if (pParent && parent == pParent->hWnd)
				home = pParent->HomeDesk;
			else
				home = GetHomeFromArea(pData);

			SetHomeDesk(pData,home);
			/* stupid "Oscar_Balloon" window... */ 
      /* (AIM hides its tooltop window by moving it to (-10,-10)-(-5,-5). */ 
      /* Not even the courtesy to size it to 0x0, let alone hide it.)     */ 
			CWTRACE(TRACE_WARNING,pData->HomeDesk<0,pData,"Found lost window.");

#if !defined(NDEBUG) && defined(TRACE_FILE) && defined(TRACE_UPDATE)
	/* to make sure it works in debug mode without flooding the logs */ 
	if (bInLocalInit) {
    TRACE(TRACE_STATUS,"-- ScreenToVWM");
	}
#endif
			/* convert to vwm coords */ 
			ScreenToVWM(&pData->Mini,&pData->Area);

			pData->hMonitor = GetWindowMonitor(pData);
		}
	}
}

// UpdateWindowListProc (for use with EnumWindows) does a lot.
//
// New windows get added to the list as they are found, and the touch
// flag for each existing window is toggled.  Windows that are removed
// are removed from the list (in UpdateWindowList below) when their
// touch flag does not get toggled.  (that is, when it matches the old
// state of the global touch flag, so that needs to be updated after
// each pass through this, in UpdateWindowList)
//
// This routine will also handle ignoring windows that the VWM can
// completely ignore, such as hidden windows and windows that are hidden
// with cheap tricks like setting their size to 0x0.  Sticky windows
// can't be ignored this way, so they are in the list.
//
// This also checks each window's position, size, and (if appropriate)
// title for changes and sets the flags if the VWM needs to be redrawn.
//
// Turning off flash when a window gains focus is also handled here.
//
// lParam is a char[MAX_LINE_LENGTH] buffer on UpdateWindowList's stack
//
static BOOL CALLBACK UpdateWindowListProc(HWND hWnd,LPARAM lParam)
{
	int i;
	s_WinData *pData;
	char *WndTitle = (char *)lParam;
	ASSERT(WRITEABLE(WndTitle,MAX_LINE_LENGTH*sizeof(char)));

#if !defined(NDEBUG) && defined(TRACE_FILE) && defined(TRACE_UPDATE)
	/* to make sure it works in debug mode without flooding the logs */ 
	if (bInLocalInit) {
		char buffer[64];
		sprintf(buffer,"->UpdateWindowListProc(%08X)",hWnd);
		TRACE(TRACE_STATUS,buffer);
	}
#endif

	if (hWnd==NULL) return TRUE;

	/* not using GetWindowPtr because we need i to insert the window in a minute */ 
	i = FindKnownWindow(hWnd);
	pData = &GlobalData.KnownWindows[i];

	if (pData->hWnd!=hWnd)
	{
		/* window not in list     */ 

		if (ShouldntAddWindow(hWnd))
			return TRUE;

#if !defined(NDEBUG) && defined(TRACE_FILE) && defined(TRACE_UPDATE)
	/* to make sure it works in debug mode without flooding the logs */ 
	if (bInLocalInit) {
    TRACE(TRACE_STATUS,"-- adding...");
	}
#endif

#ifdef DONT_WANT_PHOTOSHOP_PALETTES_TO_WORK
		if (GetWindowTextLength(hWnd)==0) /* no title */ 
			return TRUE;
#endif

		/* Add the window.  Note that it won't stay unless it's touched.     */ 
		i = InsertWindow(i,hWnd);
		pData = &GlobalData.KnownWindows[i];

		/* If the window is specifically rejected after this point,          */ 
		/* you should do a DeleteWindow(i); return TRUE;                     */ 
		/* The window would be automatically removed without the             */ 
		/* DeleteWindow(i); (so long as it's before the touch point)         */ 
		/* but it's more efficient this way.                                 */ 

	} else {
		/* window already in list */ 
		pData->bNewWindow = false;
	}

#if !defined(NDEBUG) && defined(TRACE_FILE) && defined(TRACE_UPDATE)
	/* to make sure it works in debug mode without flooding the logs */ 
	if (bInLocalInit) {
    WTRACE(TRACE_STATUS,pData,"-- Window");
	}
#endif

	/* if it's hidden, but not by me, remove it                            */ 
	/* if i've hidden it, and it wants to go away like this, er... problem */ 
	/* can't think of any solution to that though.                         */ 
	if (!IsWindowVisible(hWnd) && !pData->bHidden)
	{
#if !defined(NDEBUG) && defined(TRACE_FILE) && defined(TRACE_UPDATE)
	/* to make sure it works in debug mode without flooding the logs */ 
	if (bInLocalInit) {
    TRACE(TRACE_STATUS,"-- not visible");
	}
#endif
		if (pData->bNewWindow) DeleteWindow(i);
		return TRUE;
	}

#if !defined(NDEBUG) && defined(TRACE_FILE) && defined(TRACE_UPDATE)
	/* to make sure it works in debug mode without flooding the logs */ 
	if (bInLocalInit) {
    TRACE(TRACE_STATUS,"-- UpdateRect");
	}
#endif
	/* fixed windows need to remember their last visible position          */ 
	if (!(pData->bFix && pData->bHidden)) UpdateRect(pData);

	if (pData->Area.left==pData->Area.right
		&&pData->Area.top==pData->Area.bottom)
	{
#ifdef SHOW_BERSIRC_ON_ALL_DESKTOPS_TASK_LIST
		if (pData->bNewWindow) DeleteWindow(i);
		return TRUE; /* remove it again if it's got zero area                */ 
		/* That's the only easy way i can think of to eliminate those pesky  */ 
		/* invisible yet shown windows.  I really wish people wouldn't do    */ 
		/* that, I mean how hard is it to use ShowWindow(SW_HIDE)?           */ 
#else
    pData->bFlash = false;
#endif
	}

// touch to keep it in our list
	pData->bTouch = !pData->bTouch;
	++GlobalData.nEnumWindows;

#if !defined(NDEBUG) && defined(TRACE_FILE) && defined(TRACE_UPDATE)
	/* to make sure it works in debug mode without flooding the logs */ 
	if (bInLocalInit) {
    TRACE(TRACE_STATUS,"-- getting paint order");
	}
#endif
	/* remember order for painting                                         */ 
  /* DO NOT remove the window after this point or you'll break the chain */ 
  /* (Timothy S. of Reno Nevada broke the chain and his computer froze.) */ 
	if (pData->Prev != GlobalData.hLastEnumWnd)
	{
		CTRACE(TRACE_STATUS,!GlobalData.bChangeState,"CS: Paint order.");
		GlobalData.bChangeState = true;
		GlobalData.bChangeOrder = true;
		pData->Prev = GlobalData.hLastEnumWnd;
	}
	GlobalData.hLastEnumWnd = hWnd;

	/* if anything is or was flashing, the state is changing               */ 
	/* need to check this before and after flash changes                   */ 
	if (pData->bFlash)
  {
    TRACE(TRACE_STATUS,"bChangeState due to bFlash");
    GlobalData.bChangeState = true;
  }

	/* shouldn't get icon here because we probably don't need to update it */ 
	/* and we also don't know which icon we want, large or small           */ 
	/* and won't know until we go to draw it.  (if that even happens)      */ 

	/* might want to grab both icons here, remember to always grab if NULL */ 

#if !defined(NDEBUG) && defined(TRACE_FILE) && defined(TRACE_UPDATE)
	/* to make sure it works in debug mode without flooding the logs */ 
	if (bInLocalInit) {
    TRACE(TRACE_STATUS,"-- GetWindowText");
	}
#endif
	if (GlobalData.bDetectTitleChange || pData->bNewWindow || GlobalData.nSetupPass==2)
		GetWindowText(hWnd,WndTitle,MAX_LINE_LENGTH);

	if (GlobalData.bDetectTitleChange)
	{
		char *ptr = WndTitle;
		DWORD csum=0,t;

		/* make sure it's terminated, otherwise the while will overflow           */ 
		WndTitle[MAX_LINE_LENGTH-1]=0;

		/* calculate a checksum                                                   */ 
    /* I need to change this to a real hash algo, like trusty one-at-a-time   */ 
		while (*ptr)
		{
			t = (csum<<1); /* these three lines make it sensitive to transposition  */ 
			if (csum&0x80000000) t++; /* rotate chars that have shifted off back in */ 
			if (csum&0x00040000) t++; /* a little extra chaos                       */ 
			csum = t + *ptr++;
		}

		if (csum!=pData->TitleChecksum)
		{
			pData->TitleChecksum = csum;
			pData->bTitleChanged = true;
			/* this flag should be cleared when checked                             */ 
			/* also note that new windows will always have this flag set            */ 
			/* (unless, somehow, their title checksums to 0)                        */ 

			if (!pData->bNewWindow && GlobalData.OnTitleChange
				&& !pData->bIgnoreTitleChange)
			{
				GlobalData.OnTitleChange->pData = pData;
				AtVarExec(GlobalData.OnTitleChange,false,false);
			}
		}
		else
		{
			pData->bTitleChanged = false;
		}
	}
	if ((pData->bNewWindow
		/* first pass in setup, don't check anything */ 
		/* second pass in setup, check everything    */ 
		&& GlobalData.nSetupPass!=1)
		|| GlobalData.nSetupPass==2)
	{
		/* match against *Items      */ 
		/* this may destroy WndTitle */ 
		DoWindowMatches(pData, WndTitle);
	}

#if !defined(NDEBUG) && defined(TRACE_FILE) && defined(TRACE_UPDATE)
	/* to make sure it works in debug mode without flooding the logs */ 
	if (bInLocalInit) {
    TRACE(TRACE_STATUS,"-- OnNewWindow");
	}
#endif
	if (pData->bNewWindow
		&& GlobalData.OnNewWindow
		&& GlobalData.OnNewWindow->Hold==0) {

		GlobalData.OnNewWindow->pData = pData;
		AtVarExec(GlobalData.OnNewWindow,false,false);
		/* if they bring up a new window as a result, don't trigger again  */ 
		/* wait 1 second, 4*250 (250=timer period)                         */ 
		GlobalData.OnNewWindow->Hold = 4;
		/* I don't actually know if that works, btw. */ 

    /* gather popups to parents when requested */ 
    if (GlobalData.bKeepWindowsWithParent)
    {
      HWND hParent = GetWindow(pData->hWnd, GW_OWNER);
      s_WinData *pParent;

      if (hParent)
      {
        pParent = GetWindowPtr(hParent);
        if (pParent->hWnd == hParent && pParent->HomeDesk != -1)
          MoveWindowToDesk(pData, pParent->HomeDesk, NULL);
      }
    }
	}

#if !defined(NDEBUG) && defined(TRACE_FILE) && defined(TRACE_UPDATE)
	/* to make sure it works in debug mode without flooding the logs */ 
	if (bInLocalInit) {
    TRACE(TRACE_STATUS,"-- checking for chanes");
	}
#endif
	/* check for changes in drawn windows */ 
	if (ShouldDrawWnd(pData)) 
	{
		if (pData->bNewWindow)
		{
			CWTRACE(TRACE_STATUS,!GlobalData.bChangeState,pData,"CS: New window.");
			GlobalData.bChangeState = true;
			GlobalData.bChangeOrder = true;
		}
		else if (pData->bPosChanged || pData->bSizeChanged)
		{
			CWTRACE(TRACE_STATUS,!GlobalData.bChangeState,pData,"CS: Window moved.");
			GlobalData.bChangeState = true;
			/* execute move/resize events            */ 
			if (pData->bPosChanged)
			{
				if (GlobalData.OnMoveWindow)
				{
					GlobalData.OnMoveWindow->pData = pData;
					AtVarExec(GlobalData.OnMoveWindow,false,false);
				}
			} 
			else
			/* if it didn't move, it must've resized */ 
			{
				if (GlobalData.OnSizeWindow)
				{
					GlobalData.OnSizeWindow->pData = pData;
					AtVarExec(GlobalData.OnSizeWindow,false,false);
				}
			}
		}
	}
	else if (pData->bWasDrawn)
	{
		/* redraw if the window should vanish    */ 
		/* I don't think this should ever happen */ 
    TRACE(TRACE_STATUS,"bChangeState due to bWasDrawn");
		GlobalData.bChangeState = true;
		pData->bWasDrawn = false;
	}

#if !defined(NDEBUG) && defined(TRACE_FILE) && defined(TRACE_UPDATE)
	/* to make sure it works in debug mode without flooding the logs */ 
	if (bInLocalInit) {
    TRACE(TRACE_STATUS,"-- checking flash");
	}
#endif
	/* update window flash */ 
	if (pData->bFlash || pData->bFlashOn)
	{
		bool bSaveFlash = pData->bFlash;

		/* should it stop flashing now? */ 
		if (!pData->bFlashOn && pData->iFlashCountLeft!=0
			&& GlobalData.bFlashState && --pData->iFlashCountLeft==0)
		{
			pData->bFlash = false;
			pData->bFlashOn = true;
		}

		if (GlobalData.hRootFocusWnd==dGetAncestor(hWnd,GA_ROOTOWNER))
		{
			/* turn flash off if it has focus */ 
			pData->bFlash = false;
			pData->bFlashOn = false;
		}

		/* if anything is or was flashing, the state is changing */ 
		if (bSaveFlash) GlobalData.bChangeState = true;
	}

#if !defined(NDEBUG) && defined(TRACE_FILE) && defined(TRACE_UPDATE)
	/* to make sure it works in debug mode without flooding the logs */ 
	if (bInLocalInit) {
    TRACE(TRACE_STATUS,"-- done");
	}
#endif
	return TRUE;
}


// Add and remove windows to bring the KnownWindowList up to date.
// this toggles the global touch flag, WinData::Area is also updated
void UpdateWindowList(void)
{
	int i,last;
	char buffer[MAX_LINE_LENGTH];
	s_WinData *pData;
	HWND hWnd;
	bool bClearNewWindowHold;

#if !defined(NDEBUG) && defined(TRACE_FILE)
	char c;
#endif
#ifdef VWM_STATS
	int start = GetTickCount();
#endif

	/* handle SwitchOnAnyFocus and memory for FocusLast */ 
	hWnd = GetForegroundWindow();
	pData = GetWindowPtr(hWnd);
	if (pData->hWnd==hWnd && !ShouldntAddWindow(hWnd))
	{
		if (pData->HomeDesk!=GlobalData.CurrentDesk
      && !IsWndMinimized(pData) && !pData->bFix)
		{
			if (pData->HomeDesk < 0)
			{
        if (pData->HomeDesk == DESK_NONE)
        {
					TRACE(TRACE_STATUS,"lost window in foreground");
					if (!MoveWindowToDesk(pData,GlobalData.CurrentDesk,NULL))
            pData->HomeDesk = DESK_UNGATHERABLE;
        }
			}
			else
			{
				if (GlobalData.bSwitchOnAnyFocus
          && GlobalData.hFocusWnd!=hWnd
          && !pData->bSticky
          && !GlobalData.bOnDestroy
          )
					SwitchDesk(pData->HomeDesk);
			}
		}
		GlobalData.hFocusWnd = hWnd;
		GlobalData.hRootFocusWnd = dGetAncestor(hWnd,GA_ROOTOWNER);
	}

	/* clear hold if it was set                         */ 
	bClearNewWindowHold = (GlobalData.OnNewWindow && GlobalData.OnNewWindow->Hold);

	INITTRACE(TRACE_STATUS,"-- Enumerating windows");
	GlobalData.nEnumWindows = 0;
	GlobalData.hLastEnumWnd = NULL;
	EnumWindows(&UpdateWindowListProc,(LPARAM)buffer);
	INITTRACE(TRACE_STATUS,"-- Done enumerating windows");

  ASSERT(GlobalData.nEnumWindows<=GlobalData.nKnownWindows);
	if (GlobalData.nEnumWindows<GlobalData.nKnownWindows)
	{
		TRACE(TRACE_STATUS,"nEnumWindows < nKnownWindows");
		/* find untouched windows and remove them */ 
		pData = GlobalData.KnownWindows;
		for (i=GlobalData.nKnownWindows;i;--i)
		{
			#if !defined(NDEBUG) && defined(TRACE_FILE)
			/* new window, title changed, or removed window */ 
			/* new windows always have title changed        */ 
			if (pData->bTitleChanged
				|| (pData->bTouch==GlobalData.bWindowTouch)) {
				if (pData->bTouch!=GlobalData.bWindowTouch) {
					/* + window added             */ 
					/* = title changed            */ 
					c = pData->bNewWindow?'+':'=';
				} else {
					/* . window added and removed */ 
					/* - window removed           */ 
					c = pData->bNewWindow?'.':'-';
				}
				sprintf(buffer,"%c [%2d] = %08X { %5d, %5d, %5d, %5d } %c t#%08X %c",
					c,
					GlobalData.nKnownWindows-i,
					pData->hWnd,
					pData->Area.left,
					pData->Area.top,
					pData->Area.right,
					pData->Area.bottom,
					pData->bTitleChanged?'T':'-',
					pData->TitleChecksum,
					pData->bHidden?'H':'-'
					);
				TRACE(TRACE_STATUS,buffer);
			}
			#endif /* TRACE and debug build */ 

      /* was window touched in enumeration? */ 
			if (pData->bTouch == GlobalData.bWindowTouch) {
				if (!pData->bNewWindow) {
					if (pData->bWasDrawn) {
						CTRACE(TRACE_STATUS,!GlobalData.bChangeState,"CS: Remove Window.");
						GlobalData.bChangeState = true;
					}
					if (GlobalData.OnRemoveWindow) {
						/* only has to be valid during atvar replacement */ 
						GlobalData.OnRemoveWindow->pData = pData;
						AtVarExec(GlobalData.OnRemoveWindow,false,false);
					}
				}
				DeleteWindow(GlobalData.nKnownWindows-i);
			}
			else
			{
				++pData;
			}
		}
		GlobalData.bChangeOrder = true;
	}
#ifndef NDEBUG
	else
	{
		/* find untouched windows and remove them */ 
		pData = GlobalData.KnownWindows;
		for (i=GlobalData.nKnownWindows;i;--i)
		{
			#if !defined(NDEBUG) && defined(TRACE_FILE)
			/* new window, title changed, or removed window */ 
			/* new windows always have title changed        */ 
			if (pData->bTitleChanged
				|| (pData->bTouch==GlobalData.bWindowTouch)) {
				if (pData->bTouch!=GlobalData.bWindowTouch) {
					/* + window added             */ 
					/* = title changed            */ 
					c = pData->bNewWindow?'+':'=';
				} else {
					/* . window added and removed */ 
					/* - window removed           */ 
					c = pData->bNewWindow?'.':'-';
				}
				sprintf(buffer,"%c [%2d] = %08X { %5d, %5d, %5d, %5d } %c t#%08X %c",
					c,
					GlobalData.nKnownWindows-i,
					pData->hWnd,
					pData->Area.left,
					pData->Area.top,
					pData->Area.right,
					pData->Area.bottom,
					pData->bTitleChanged?'T':'-',
					pData->TitleChecksum,
					pData->bHidden?'H':'-'
					);
				TRACE(TRACE_ERROR,"Untouched window left in list.");
				TRACE(TRACE_ERROR,buffer);
			}
			#endif
			if (pData->bTouch == GlobalData.bWindowTouch) {
				if (!pData->bNewWindow) {
					if (pData->bWasDrawn) {
						CTRACE(TRACE_STATUS,!GlobalData.bChangeState,"CS: Remove Window.");
						GlobalData.bChangeState = true;
					}
					if (GlobalData.OnRemoveWindow) {
						/* only has to be valid during atvar replacement */ 
						GlobalData.OnRemoveWindow->pData = pData;
						AtVarExec(GlobalData.OnRemoveWindow,false,false);
					}
				}
				DeleteWindow(GlobalData.nKnownWindows-i);
			}
			else
			{
				++pData;
			}
		}
	}
#endif /* debug build */ 

	ASSERTV(GlobalData.nEnumWindows==GlobalData.nKnownWindows);

	/* bring global touch into sync */ 
	GlobalData.bWindowTouch = !GlobalData.bWindowTouch;

	/* Find window paint order.                         */ 
	/* Conveniently, windows are enumerated from top to */ 
	/* bottom, and we want to paint from bottom to top. */ 
  if (GlobalData.bChangeOrder)
	{
	  /* these are only for ChangeOrder false alarm */ 
	  /* detection, so only do it in debug builds   */ 
#ifndef NDEBUG
    GlobalData.bChangeOrder = false;
#define CHECK_CHANGEORDER(n) if (!(n)) GlobalData.bChangeOrder = true
#else
#define CHECK_CHANGEORDER(n) (void)0
#endif

		TRACE(TRACE_STATUS,"Getting paint order");
		/* We do this now because it's the point where we first  */ 
		/* know each window's final position in the KnownWindows */  
		/* array, so we couldn't get proper pointers before.     */ 
		if (GlobalData.nKnownWindows)
		{
			ASSERTV(GlobalData.hLastEnumWnd!=NULL);
			/* this will succeed if there's at least one wnd */ 
			/* so we don't need to check the hWnd            */ 
			i = FindKnownWindow(GlobalData.hLastEnumWnd);
			ASSERTV(GlobalData.KnownWindows[i].hWnd == GlobalData.hLastEnumWnd);

      CHECK_CHANGEORDER(GlobalData.FirstWindow == i);
			GlobalData.FirstWindow = last = i;

			while (GlobalData.KnownWindows[last].Prev!=NULL)
			{
				i = FindKnownWindow(GlobalData.KnownWindows[last].Prev);
        /* we just updated this, so the window should */ 
        /* still be there, no need to check           */ 
				ASSERTR(GlobalData.KnownWindows[last].Prev==GlobalData.KnownWindows[i].hWnd,;);
        ASSERTV(last!=i);

        CHECK_CHANGEORDER(GlobalData.KnownWindows[last].PaintNext == i);
        GlobalData.KnownWindows[last].PaintNext = i;

				last = i;
			}

      CHECK_CHANGEORDER(GlobalData.KnownWindows[last].PaintNext == -1);
			GlobalData.KnownWindows[last].PaintNext = -1;
		}
		else
		{
			/* no windows to draw */ 
      CHECK_CHANGEORDER(GlobalData.FirstWindow == -1);
			GlobalData.FirstWindow = -1;
		}

		/* this actually happens, but it's only a       */ 
		/* performance issue (and a slight one at that) */ 
		/* would be nice to fix it at some point.       */ 
		CTRACE(TRACE_ERROR,GlobalData.bChangeOrder,"bChangeOrder: false alarm.");
    TRACE(TRACE_STATUS,"bChangeState due to bChangeOrder");
#undef CHECK_CHANGEORDER

		GlobalData.bChangeState = true;
		GlobalData.bChangeOrder = false;
	}

	/* window sizes/positions have been updated */ 
	GlobalData.bScaleChanged = false;

	if (bClearNewWindowHold) GlobalData.OnNewWindow->Hold--;

	// if a new window stole focus and is on a different desk, reject it
	/* This doesn't entirely work, and sometimes it prevents windows from gaining
	   focus when they actually should.  What I really need is some way to prevent
		 focus from changing in the first place, but oh well.
	pData = GetWindowPtr(hWnd);
	if (pData->hWnd==hWnd
			&& pData->bNewWindow
			&& pData->HomeDesk!=GlobalData.CurrentDesk)
		SetDeskFocus();
		*/

#if !defined(NDEBUG) && defined(TRACE_FILE)
	if (bInLocalInit) DumpWindowList();
#endif
#ifdef VWM_STATS
	VWMStats.UpdateTime += GetTickCount()-start;
#endif
}

// move a window
// desk = the desk that will be current when we're done
//        or the current one if we're not switching desks
// wndesk = the desk that the window will be on
// hDwp = DeferWindowPos struct, or NULL to move now
bool MoveWindowVWM(s_WinData *pData, const SIZE *amount, int desk, int wndesk, HDWP hDwp)
{
	ASSERTR(READABLE(pData,sizeof(s_WinData)),false);
	ASSERTR(WRITEABLE(pData,sizeof(s_WinData)),false);
	ASSERTR(READABLE(amount,sizeof(SIZE)),false);

	if (pData->bSticky) return true;

	if ((ShouldHideWnd(pData) && desk==wndesk)
			|| GlobalData.bUnHideTaskOnSwitch)
  {
		TRACE(TRACE_STATUS,"Showing.");
		CWTRACE(TRACE_WARNING,IsWndHung(pData->hWnd),pData,"Hung window: showing asynchronously.");
		/* hmm, with async windows don't always repaint */ 
		/* without async, external windows could freeze */ 
		/* the module, and all of litestep              */ 
		if (IsWndHung(pData->hWnd))
			ShowWindowAsync(pData->hWnd,SW_SHOWNA);
		else ShowWindow(pData->hWnd,SW_SHOWNA);
		pData->bHidden = false;
		/* -> should use home desk of owner window */ 
		/* showing, move back to original position */ 
	}
	if (ShouldMoveWnd(pData))
  {
		CWTRACE(TRACE_WARNING,IsWndHung(pData->hWnd),pData,"Hung window: moving asynchronously.");
		if (hDwp!=NULL &&
			/* if we know it won't respond, move it later    */ 
			/* rather than causing the whole thing to fail   */ 
			!IsWndHung(pData->hWnd)
			) {
			/* mark the window to be moved                   */ 
			DeferWindowPos(hDwp,
				pData->hWnd,
				HWND_TOP,         /* hWndInsertAfter           */ 
				pData->Area.left+amount->cx,
				pData->Area.top +amount->cy,
				0, 0,							/* cx, cy                    */ 
        SWP_NOZORDER |		/* ignore hWndInsertAfter    */ 
				SWP_NOSIZE |			/* ignore cx and cy          */ 
				SWP_NOACTIVATE
				);
		} else {
			/* move the window now (asynchronously, that is) */ 
			SetWindowPos(
				pData->hWnd,
				HWND_TOP,
				pData->Area.left+amount->cx,
				pData->Area.top +amount->cy,
				0, 0,
				SWP_NOZORDER | SWP_NOSIZE | SWP_NOACTIVATE |
				SWP_ASYNCWINDOWPOS /* to stop freezing ls with dead windows */ 
				/* actually, try it!  it's cool to watch the window dance   */ 
				/* across the screen all by themselves!                     */ 
				);
		}
	}
	if (ShouldHideWnd(pData) && desk!=wndesk)
  {
		TRACE(TRACE_STATUS,"Hiding.");
		ShowWindowAsync(pData->hWnd,SW_HIDE);
		pData->bHidden = true;
	}

	/* Don't update the s_WinData rectangles, let the update fxn   */ 
	/* get that.  Otherwise updates don't always happen on moves.  */ 

	/* should move any top-level children this window has as well  */ 
	/* maybe move all windows with the same ROOTOWNER?             */ 
	/* erg, that would be slow (or would it?)                      */ 

	/* should to make this return false if the move function fails */ 
	return true;
}

struct MoveEachWindow
{
  const SIZE *amount;
  int iEndingCurrentDesk;
  HDWP hDwp;

  MoveEachWindow(const SIZE *amount_, int desk, HDWP hDwp_)
    : amount(amount_), iEndingCurrentDesk(desk), hDwp(hDwp_)
  { }

  bool operator()(s_WinData *pData)
  {
		if (pData->hWnd != GlobalData.hWndDragging)
    {
      int wnddesk = pData->HomeDesk;
      /* keep fixed windows with their parent */ 
      if (ShouldFixWnd(pData))
      {
        HWND hParent = GetParent(pData->hWnd);
        s_WinData *pParent = GetWindowPtr(hParent);
        if (pParent->hWnd == hParent)
          wnddesk = pParent->HomeDesk;
      }
			MoveWindowVWM(pData,amount,iEndingCurrentDesk,wnddesk,hDwp);
    }
		else
			/* make sure the miniwnd moves to the right place */ 
			/* when dragging a window while switching desks   */ 
			pData->bPosChanged = true;

    return false;
  }
};

// moves all known windows by amount
// desk specifies which desk we're moving TO
bool MoveAllWindows(const SIZE *amount, int desk)
{
	HDWP hDwp;

	ASSERTR(READABLE(amount,sizeof(SIZE)),false);
	ASSERTR(desk>=0 && desk<GlobalData.nDesks,false);

	/* tell Windows we're moving a lot of stuff */ 
	hDwp = BeginDeferWindowPos(GlobalData.nKnownWindows);

	/* fail if not enough memory                */ 
	/* (actually, we could move them anyway)    */ 
	if (hDwp==NULL)
	{
		TRACE(TRACE_WARNING,"Not enough memory for DeferWindowPos.");
		return false;
	}

  MoveEachWindow proc(amount, desk, hDwp);
  IterateAnyOrder(proc);

	/* move them all                            */ 
	/* EndDeferWindowPos returns nonzero on success, translate that to true */ 
	return !!EndDeferWindowPos(hDwp);
}

// #acdpht
DWORD ParseWindowSpecFlags(const char *spec)
{
	DWORD flags = 3;
	ASSERTR(spec!=NULL,flags);

	if (*spec=='#') {
		flags = 0;
		while (*++spec) {
			switch (*spec) {
			case 'a':
			case 'A': /* match more than one window      */ 
        /* this doesn't have meaning for *lines of course */ 
				TRACE(TRACE_STATUS,"MATCH_WND_ALL");
				flags |= MATCH_WND_ALL;
				break;
			case 'c':
			case 'C': /* match window class              */ 
				TRACE(TRACE_STATUS,"MATCH_WND_CLASS");
				flags |= MATCH_WND_CLASS;
				break;
			case 'd':
			case 'D': /* match on current desk only      */ 
				TRACE(TRACE_STATUS,"MATCH_WND_DESKONLY");
				flags |= MATCH_WND_DESKONLY;
        /* does it have a desk #? */ 
				if (spec[1]>='0' && spec[1]<='9') {
					++spec;
					GlobalData.MatchDesk = atoi(spec)-1;
					while (*spec>='0' && *spec<='9') {
						spec++;
						if (*spec=='.') /* anything after a dot is a y coord */ 
							GlobalData.MatchDesk += (atoi(++spec)-1)*GlobalData.nDesksX;
					}
					--spec;
					if (GlobalData.MatchDesk<0) GlobalData.MatchDesk = GlobalData.CurrentDesk;
				}
				else
				{
					GlobalData.MatchDesk = GlobalData.CurrentDesk;
				}
				break;
			case 'h':
			case 'H': /* can't use other flags with this */ 
				TRACE(TRACE_STATUS,"MATCH_WND_BYHANDLE");
				return MATCH_WND_BYHANDLE;
			case 'p':
			case 'P': /* match by paint order            */ 
				TRACE(TRACE_STATUS,"MATCH_WND_PAINTORDER");
				flags |= MATCH_WND_PAINTORDER;
				break;
			case 't':
			case 'T': /* match window title              */ 
				TRACE(TRACE_STATUS,"MATCH_WND_TITLE");
				flags |= MATCH_WND_TITLE;
				break;
			case ' ':
			case '\t':
				return flags;
			}
		}
	}
	return flags;
}

// "Does this window match this windowspec?"
// don't I do this someplace else?  should merge them.
bool CheckMatchWindow(const s_WinData *pData,const char *spec,DWORD flags)
{
	char buffer[256];
	ASSERTR(READABLE(pData,sizeof(s_WinData)),false);
	ASSERTR(READABLE(spec,sizeof(char)),false);

	if (flags&MATCH_WND_DESKONLY && pData->HomeDesk != GlobalData.MatchDesk)
		return false;

	/* let me match everything without specifying a class/title */ 
	/* but still let me match an empty title specifically       */ 
	/* I think it's fairly safe to assume no NULL window class  */ 
	if (flags&MATCH_WND_CLASS && spec==NULL || *spec==0) return true;

	if (flags&MATCH_WND_TITLE) {
		GetWindowText(pData->hWnd, buffer, 256);
		if (match(spec,buffer)) return true;
	}
	if (flags&MATCH_WND_CLASS) {
		GetClassName(pData->hWnd, buffer, 256);
		if (match(spec,buffer)) return true;
	}
	return false;
}

// dummy WinModifier proc to just return the window
void GetWindowProc(s_WinData *pData,DWORD ptr)
{
	ASSERTR(ptr!=0,;);
	*(s_WinData **)ptr = pData;
}

struct ExecWinModifier
{
  DWORD flags;
  const char *spec;
  WinModifier *func;
  DWORD lParam;

  ExecWinModifier(DWORD flags_, const char *spec_, WinModifier *func_, DWORD lParam_)
    : flags(flags_), spec(spec_), func(func_), lParam(lParam_)
  { }

  bool operator()(s_WinData *pData)
  {
		if (CheckMatchWindow(pData,spec,flags))
    {
			/* found one */ 
			func(pData,lParam);
      DUMPWIN(TRACE_STATUS,pData);
      return !(flags&MATCH_WND_ALL);
		}
    return false;
  }
};

// Execute WinModifier function for all windows matching spec
// if spec is an empty string, the last focused window alone is used
void ExecWindowBySpec(const char *spec, WinModifier *func, DWORD lParam)
{
	s_WinData *pData;
	DWORD flags = MATCH_WND_TITLE|MATCH_WND_CLASS;
	HWND hWnd = GlobalData.hFocusWnd;

	TRACE(TRACE_STATUS,"ExecWindowBySpec");
	TRACE(TRACE_STATUS,spec);

	ASSERTR(READABLE(spec,sizeof(char)),;);
	ASSERTR(EXECABLE(func),;);

	if (*spec==0)
	{
		/* if no window spec given, use last focus window */ 
		pData = GetWindowPtr(hWnd);
		if (pData->hWnd == hWnd)
    {
			func(pData,lParam);
      DUMPWIN(TRACE_STATUS,pData);
		}
	}
	else
	{
		/* see if the spec includes flags */ 
		if (*spec=='#')
    {
			flags = ParseWindowSpecFlags(spec);
			while (*spec!=' ' && *spec!=0) spec++;
			while (*spec==' ') spec++;
		}

    ExecWinModifier ewm(flags, spec, func, lParam);

		if (flags&MATCH_WND_SPECIAL)
    {
			/* special match type (doesn't use other flags) */ 
			switch (flags)
      {
			case MATCH_WND_BYHANDLE:
				/* allow hex starting with 0x */ 
				if (*spec=='0')
				{
					++spec;
					if (*spec=='x' || *spec=='X')
					{
						++spec;
						sscanf(spec,"%x",&hWnd);
						break;
					}
				}
				/* window handle in decimal */ 
				sscanf(spec,"%d",&hWnd);
				break;
			default:
				TRACE(TRACE_ERROR,"Unknown special window spec.");
			}
			if (hWnd!=NULL) /* fail silently for null window */ 
      {
				pData = GetWindowPtr(hWnd);
				if (pData->hWnd==hWnd)
        {
					func(pData,lParam);
          DUMPWIN(TRACE_STATUS,pData);
					/* only one can match */ 
				}
			}
		}
    else if (flags&MATCH_WND_PAINTORDER)
      IteratePaintOrder(ewm);
    else
      IterateAnyOrder(ewm);
	}
}

// Move the center of the window that pData refers to to desk nDesk
// nDesk must be a valid, 0-based desk number
// hDwp can be NULL to move now rather than deferred.
bool MoveWindowToDesk(s_WinData *pData,int nDesk,HDWP hDwp)
{
	SIZE dif;

	ASSERTR(READABLE(pData,sizeof(s_WinData)),false);

	WTRACE(TRACE_STATUS,pData,"MoveWindowToDesk");

	/* don't move to an invalid desk  */ 
	if (nDesk<0 || nDesk>=GlobalData.nDesks) return false;

	/* get center                     */ 
	dif.cx = (pData->Area.left + pData->Area.right)/2;
	dif.cy = (pData->Area.top + pData->Area.bottom)/2;

	DeskFromPointXY(dif.cx,dif.cy,(POINT *)&dif);

	/* need offset from here to there */ 
	GetDeskDifXY(
		nDesk%GlobalData.nDesksX,
		nDesk/GlobalData.nDesksX,
		dif.cx,dif.cy,&dif);

	return MoveWindowVWM(pData,&dif,GlobalData.CurrentDesk,nDesk,hDwp);
}

struct GatherWindow
{
  int desk;
  HDWP hDwp;

  GatherWindow(int desk_, HDWP hDwp_) : desk(desk_), hDwp(hDwp_) { }

  bool operator()(s_WinData *pData)
  {
		if (!IsWndMinimized(pData)) {
			MoveWindowToDesk(pData,desk,hDwp);
			if (pData->bHidden) {
				/* make sure they get shown */ 
				ShowWindow(pData->hWnd,SW_SHOWNA);
				pData->bHidden = false;
			}
		}
    return false;
  }
};

// gather all windows to desk
// desk must be a valid, 0-based desk number
bool Gather(int desk)
{
	HDWP hDwp;

	/* don't move to an invalid desk            */ 
	if (desk<0 || desk>=GlobalData.nDesks) return false;

	/* tell Windows we're moving a lot of stuff */ 
	hDwp = BeginDeferWindowPos(GlobalData.nKnownWindows);

	/* fail if not enough memory                */ 
	/* (actually, we could move them anyway)    */ 
	if (hDwp==NULL) return false;

  GatherWindow gw(desk, hDwp);
  IterateAnyOrder(gw);

	/* move them all */ 
	/* EndDeferWindowPos returns nonzero on success, translate that to true */ 
	return !!EndDeferWindowPos(hDwp);
}

struct GatherLostWindow
{
  int desk;
  bool bAllOk;

	HWND parent;
  s_WinData *pParent;

  GatherLostWindow(int desk_) : desk(desk_), bAllOk(true) { }

  bool operator()(s_WinData *pData)
  {
		if (pData->HomeDesk<0 && !IsWndMinimized(pData))
    {
			if (pData->bHidden && pData->bFix)
      {
				/* fixed windows should be gatherred to their parents instead */ 
				parent = GetParent(pData->hWnd);
				pParent = GetWindowPtr(parent);

        if (pParent->hWnd == parent)
        {
				  SetHomeDesk(pData,pParent->HomeDesk);
				  bAllOk = MoveWindowToDesk(pData,pData->HomeDesk,NULL) && bAllOk;
        }
        else
				  bAllOk = MoveWindowToDesk(pData,desk,NULL) && bAllOk;
			}
      else
      {

				bAllOk = MoveWindowToDesk(pData,desk,NULL) && bAllOk;

				if (pData->bHidden)
        {
					/* make sure they get shown */ 
					ShowWindow(pData->hWnd,SW_SHOWNA);
					pData->bHidden = false;
				}
			}
		}
    return false;
  }
};

// gather window that are outside the VWM area
bool GatherLost(int desk)
{
#if !defined(NDEBUG) && defined(TRACE_FILE)
	{
		char buffer[64];
		sprintf(buffer,"Gathering lost to %d.",desk);
		TRACE(TRACE_STATUS,buffer);
	}
#endif

	/* don't move to an invalid desk  */ 
	if (desk<0 || desk>=GlobalData.nDesks) return false;

  GatherLostWindow gwl(desk);
  IterateAnyOrder(gwl);

	CTRACE(TRACE_WARNING,!gwl.bAllOk,"failed.");

	return gwl.bAllOk;
}

struct MatchWindowByVWMPoint
{
  const POINT *pt;
  s_WinData *pFound;

  MatchWindowByVWMPoint(const POINT *pt_) : pt(pt_), pFound(NULL) { }

  bool operator()(s_WinData *pData)
  {
		if (ShouldDrawWnd(pData) && PtInRect(&pData->Mini,*pt))
			pFound = pData; /* save it and keep going, this is back to front */ 
    return false;
  }
};

// find the topmost window at this point
s_WinData *WindowFromVWMPoint(const POINT *pt)
{
  MatchWindowByVWMPoint mw(pt);
  IteratePaintOrder(mw);
  return mw.pFound;
}

// eat downtown tokyo
void SetWndAlpha(s_WinData *pData, int alpha)
{
  ASSERT(WRITEABLE(pData,sizeof(s_WinData)));
	if (alpha!=(alpha&255)) alpha = GlobalData.cWnd.Alpha;
	pData->Alpha = alpha;
  TRACE(TRACE_STATUS,"bChangeState due to SetWndAlpha");
	GlobalData.bChangeState = true;
}

// Emulate GetAncestor for windows 95
// We only ever use it for GA_ROOTOWNER, so this should be enough.
HWND WINAPI GetAncestor95(HWND hWnd, UINT rootOwner)
{
	HWND last;

	ASSERTV(rootOwner==GA_ROOTOWNER);

	while (hWnd!=NULL)
	{
		last = hWnd;
		hWnd = GetParent(last);
	}
	return last;
}

// Set the VWM's position, taking into account ontopness and whether or not it's a child window.
// That entails fixing up the coords for negative-space monitors in some cases.
void SetVWMPos(void)
{
	if (GlobalData.bAlwaysOnTop
    || GlobalData.hParentWnd==NULL
    || GlobalData.bStayChildWindow)
		/* 0,0 is origin of primary desk  */ 
		SetWindowPos(GlobalData.hMainWnd,NULL,
		  GlobalData.VWMPos.x,GlobalData.VWMPos.y,
		  0,0,SWP_NOZORDER|SWP_NOSIZE|SWP_NOACTIVATE);
	else
		/* 0,0 is origin of parent window */ 
		SetWindowPos(GlobalData.hMainWnd,NULL,
		  GlobalData.VWMPos.x-GlobalData.ParentRect.left,
		  GlobalData.VWMPos.y-GlobalData.ParentRect.top,
		  0,0,SWP_NOZORDER|SWP_NOSIZE|SWP_NOACTIVATE);
}
